#!usr/bin/env python
# -*- coding: utf8 -*-

# -----------------------------------------------------------------------------
#   File: import_annotations.py (as part of project URUMETRICS)
#   Created: 23/05/2022 11:28
#   Last Modified: 23/05/2022 11:28
# -----------------------------------------------------------------------------
#   Author: William N. Havard
#           Postdoctoral Researcher
#
#   Mail  : william.havard@ens.fr / william.havard@gmail.com
#  
#   Institution: ENS / Laboratoire de Sciences Cognitives et Psycholinguistique
#
# ------------------------------------------------------------------------------
#   Description: 
#       • 
# -----------------------------------------------------------------------------

# !usr/bin/env python
# -*- coding: utf8 -*-

#
# Author: William N. Havard (base on various files by Lucas Gautheron)
#

import logging
import os

import pandas as pd
from ChildProject.annotations import AnnotationManager
from ChildProject.projects import ChildProject

from consts import ANNOTATION_TYPES
from utils import get_raw_filename
from custom_converters import FilteredCsvConverter # /!\ Do not remove this line

logger = logging.getLogger(__name__)
pd.set_option('mode.chained_assignment', None)  # Silences pandas' complaints


def _filter_missing_annotation_files(path_project, input):
    """
    from the dataframe given as input, Checks whether some raw annotation files found in column
    'raw_filename' are missing in the dataset.
    Returns only the rows for which the raw annotation file could be found in the dataset.
    :param path_project: path to ChildProject project directory
    :type path_project: str
    :param input: dataframe containing the list of recordings and their corresponding annotation file, has the format of 'annotations.csv'
    :type input: pandas.DataFrame
    :return: dataframe containing only row for which an annotation file was found
    :rtype: pandas.DataFrame
    """
    annotation_path = lambda row: os.path.join(path_project, 'annotations', row['set'], 'raw', row['raw_filename'])
    input['exists'] = input.apply(lambda row: os.path.isfile(annotation_path(row)), axis=1)

    missing_annotations = input[input['exists'] == False]
    existing_annotations = input[input['exists'] == True]

    if len(missing_annotations):
        missing_annotations['expected_path'] = missing_annotations.apply(lambda row: annotation_path(row), axis=1)
        missing_annotations_path = sorted(missing_annotations['expected_path'].tolist())
        missing_message = "Some annotations you expected to have are missing.\n" \
                          "Check whether these annotations exist and if so, if their expected path " \
                          "reflect their true path.\n\t - {}".format('\n\t - '.join(missing_annotations_path))
        logger.warning(missing_message)
    return existing_annotations


def _check_importation(am, imported, expected_ann_number):
    """
    Checks whether the importation was carried out successfully or not. Returns the number of imported annotations files
    imported so far and the number of segments they contain.
    :param am: ChildProject annotation manager object
    :type am: ChildProject.AnnotationManager
    :param imported: DataFrame object containing the annotation files that were successfully imported
    :type imported: pandas.DataFrame
    :param expected_ann_number: containing the expected number of annotation files that should have been imported
    :type expected_ann_number: int
    :return: number of imported annotations, number of imported segments
    :rtype: tuple of int
    """
    if len(imported) != expected_ann_number:
        logger.warning('Expected to import {} annotations, only found {}!'.format(len(imported), expected_ann_number))

    annotations_segments = am.get_segments(imported)
    if len(annotations_segments) == 0:
        logger.warning('Annotations were imported, but they either contain no segments '
                       'or the segments were not imported properly!')

    return len(imported), len(annotations_segments)


def _get_recordings(project, annotation_set, annotation_format, recording):
    """
    Returns a DataFrame of recordings already imported for the project `project` and prepare the data frame indicating
    the format and the set of annotations that will be imported.
    :param project: ChildProject project object
    :type project: ChildProject.projects.ChildProject
    :param annotation_set: set of annotation to import
    :type annotation_set: str
    :param annotation_format: format of the annotations that will be imported
    :type annotation_format: str
    :return: dataframe containing the recordings found for the project
    :rtype: pandas.DataFrame
    """
    input = project.recordings[['recording_filename', 'duration', 'child_id']]
    input.dropna(inplace=True)
    input = input[input['recording_filename'] != 'NA']
    input['set'] = annotation_set
    input['format'] = annotation_format
    input['time_seek'] = 0
    input['range_onset'] = 0
    input['range_offset'] = input['duration']
    input.drop(['duration'], axis=1, inplace=True)

    if recording:
        input = input[input['recording_filename'] == recording]

    return input


def _build_raw_filename(input, annotation_set, filename='', extension=''):
    """
    Build the expected annotation filename path containing the annotation corresponding to a given recording
    :param input: DataFrame containing a list of recordings
    :type input: pandas.DataFrame
    :param annotation_set: annotations set to import to, must be in the list of supported types
    :type annotation_set: str
    :param filename: filename of the annotation file where to look for the annotations for a specific recording
    :type filename: str
    :param extension: file extension of the annotation file, defaults to the extension found for that type
    :type extension: str
    :return: dataframe containing the expected name the annotation file should have
    :rtype: pandas.DataFrame
    """
    # Set up 'raw_filename' and 'filter' depending on the annotation set to import
    annotation_format_extension = ANNOTATION_TYPES.asdict()[annotation_set.upper()][-1] \
        if not extension else extension
    if annotation_set in ['vtc', 'vcm', 'alice', 'acoustic', 'conversations']:
        # Annotations have the same name as the recording filename. Update if it's not the case for you.
        input['raw_filename'] = input['recording_filename'].apply(
            lambda f: '{}.{}'.format(get_raw_filename(f), annotation_format_extension)) if not filename else filename
        # We only keep lines for which the 'file' column is equal to filter
        input['filter'] = input['recording_filename'].apply(lambda f: os.path.basename(get_raw_filename(f)))
    elif annotation_set in ['cha', 'its', 'eaf']:
        # CHA/ITS/EAF files do not need filtering as they only contain annotations for the file they are linked to
        input['raw_filename'] = input['recording_filename'].apply(
            lambda f: '{}.{}'.format(get_raw_filename(f), annotation_format_extension))
    else:
        raise ValueError('Unknown annotation format for set `{}`!'.format(annotation_set))

    return input


def _import_annotation(project, am, annotation_type, annotation_file, recordings_from_annotation_file, recording):
    """
    import annotations into the dataset, using a defined type and annotation raw file. 
    The type induces the set and how to convert the raw file
    :param project: ChildProject project object
    :type project: ChildProject.projects.ChildProject
    :param am: Annotation manager for this project
    :type am: ChildProject.annotations.AnnotationManager
    :param annotation_type: Type of the annotation to import, will induce the set name and format (converter), must exist in import_data.consts.ANNOTATION_TYPES
    :type annotation_type: str
    :param annotation_file: name of the raw file to import (situated somwhere in annotation/<set>/raw)
    :type annotation_file: str
    :param recordings_from_annotation_file: specify a raw imported annotation file, the recording it is linked to will be the only one for which new annotation can be imported, this allows to only keep one rec when multiple are in the raw input
    :type recordings_from_annotation_file: str
    :param recording: only keep import for one recording (inside recordings.csv as 'recording_filename')
    :type recording: str
    """
    annotation_set = annotation_type.lower()
    annotation_format = ANNOTATION_TYPES.asdict()[annotation_set.upper()][1]

    # Get recordings and set up df
    input = _get_recordings(project, annotation_set, annotation_format, recording)

    # Build raw file names
    input = _build_raw_filename(input, annotation_set, annotation_file)

    if recordings_from_annotation_file:
        anns_from_annotation_file = am.annotations[am.annotations['raw_filename'] == recordings_from_annotation_file]
        input = input[input['recording_filename'].isin(anns_from_annotation_file['recording_filename'])]

    # Filter out rows for which we do not find the matching annotation file
    input = _filter_missing_annotation_files(project.path, input)

    # We make sure we remove annotation files from the DataFrame if they were already previously imported
    if len(am.annotations):
        already_imported_metadata = am.annotations
        already_imported_metadata = already_imported_metadata[['recording_filename', 'set']]

        # Drop annotation files that were already imported
        input = (input.merge(already_imported_metadata, how='left', indicator=True,
                             on=['recording_filename', 'set'])
                 .loc[lambda x: x['_merge'] == 'left_only']
                 .drop(columns="_merge"))

    # Do importation
    if len(input) > 0:
        imported = am.import_annotations(input, threads=1)

        expected_ann_number = len(input)
        len_ann, len_seg = _check_importation(am, imported, expected_ann_number)
        logger.info('Imported {} new annotation files resulting '
                    'in {} new {} segments!'.format(len_ann, len_seg, annotation_type))

    else:
        logger.warning('Nothing to import for annotation type {}!'.format(annotation_format))



def import_annotations(project_path, annotation_type, annotation_file, recordings_from_annotation_file, recording):
    """
    import annotation for a path to project, specific raw file and type, can filter recordings considered base on another
    raw filename or directly a recording name
    :param project: ChildProject project object
    :type project: ChildProject.projects.ChildProject
    :param annotation_type: Type of the annotation to import, will induce the set name and format (converter), must exist in import_data.consts.ANNOTATION_TYPES
    :type annotation_type: str
    :param annotation_file: name of the raw file to import (situated somwhere in annotation/<set>/raw)
    :type annotation_file: str
    :param recordings_from_annotation_file: specify a raw imported annotation file, the recording it is linked to will be the only one for which new annotation can be imported, this allows to only keep one rec when multiple are in the raw input
    :type recordings_from_annotation_file: str
    :param recording: only keep import for one recording (inside recordings.csv as 'recording_filename')
    :type recording: str
    """

    # Load project
    project = ChildProject(project_path)
    am = AnnotationManager(project)
    am.read()

    _import_annotation(project, am, annotation_type, annotation_file, recordings_from_annotation_file, recording)


def main(project_path, annotation_type, annotation_file, recordings_from_annotation_file= None,recording=None):
    # Check if running the script from the root of the data set
    expected_annotation_path = os.path.join(project_path, 'annotations')

    assert os.path.exists(expected_annotation_path), \
        ValueError('Expected annotation ({}) path not found. Are you sure to be running this '
                   'command from the root of the data set?'.format(expected_annotation_path))

    assert not(recording and recordings_from_annotation_file), \
        ValueError('--recording and  --recordings-from-annotation-file cannot be used together!')


    import_annotations(project_path, annotation_type, annotation_file, recordings_from_annotation_file, recording)


def _parse_args(argv):
    import argparse

    parser = argparse.ArgumentParser(description='Import annotations.')
    parser.add_argument('--project-path', required=False, type=str, default='',
                        help="Path to a ChildProject/datalad project (useful for debugging purposes).")
    parser.add_argument('--annotation-type', required=True, choices=ANNOTATION_TYPES.asdict().keys(),
                        help='Type of annotation to be imported.')
    parser.add_argument('--annotation-file', required=True,
                        help='Annotation file to be imported.')
    parser.add_argument('--recording', required=False,
                        help='If only the annotations for one specific recording should be imported, specify the name '
                             'here. Otherwise, all recordings will be considered.')
    parser.add_argument('--recordings-from-annotation-file', required=False,
                        help='Only import annotations of --annotation-file for all the recordings that were annotated '
                             'in the file --recordings-from-annotation-file')
    args = parser.parse_args(argv)

    return vars(args)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    import sys
    pgrm_name, argv = sys.argv[0], sys.argv[1:]
    args = _parse_args(argv)

    try:
        main(**args)
        sys.exit(0)
    except Exception as e:
        logger.exception(e)
        sys.exit(1)
